/**
 * All content on this website (including text, images, source
 * code and any other original works), unless otherwise noted,
 * is licensed under a Creative Commons License.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) Open-Xchange Inc., 2006-2012
 * Mail: info@open-xchange.com
 *
 * @author Stefan Eckert <stefan.eckert@open-xchange.com>
 */

define('io.ox/office/drawinglayer/view/chartstyleutil',
        ['io.ox/office/tk/utils',
         'io.ox/office/editframework/model/format/color',
         'io.ox/office/spreadsheet/utils/clipboard'], function (Utils, Color, Clipboard) {

    'use strict';

    var BACKGROUNDTRANSFORMATION = [{type: 'satMod', value: 25000}, {type: 'lumOff', value: 35000}];

    var STYLESET = [{
        bg: {type: 'scheme', value: 'light1'},
        bevelEnabled: false
    }, {
        bg: {type: 'scheme', value: 'light1'},
        bevelEnabled: false
    }, {
        bg: {type: 'scheme', value: 'light1'},
        bevelEnabled: false
    }, {
        bg: {type: 'scheme', value: 'light1'},
        bevelEnabled: true
    }, {
        bg: null,
        bevelEnabled: false
    }, {
        bg: {type: 'scheme', value: 'background1', transformations: [{type: 'lumOff', value: -80000}]},
        bevelEnabled: true
    }];

    var COLORSET = [];

    (function () {
        var allAccents = [];
        for (var i = 1; i <= 6; i++) {
            allAccents.push({type: 'scheme', value: 'accent' + i});
        }
        COLORSET.push({ name: 'Greyscale', colors: [{type: 'rgb', value: '555555'}, {type: 'rgb', value: '9E9E9E'}, {type: 'rgb', value: '727272'}, {type: 'rgb', value: '464646'}, {type: 'rgb', value: '838383'}, {type: 'rgb', value: 'C1C1C1'}], variations: []});
        COLORSET.push({ name: 'Accents', colors: allAccents, variations: [{transformations: [{type: 'lumOff', value: -1000}]}]});



        var accentVariations = [];//[{transformations: [{type: 'lumOff', value: -10000}]}, null, {transformations: [{type: 'lumOff', value: 10000}]}, {transformations: [{type: 'lumOff', value: 20000}]}, {transformations: [{type: 'lumOff', value: 30000}]}, {transformations: [{type: 'lumOff', value: 40000}]}];

        for (var i = -2; i <= 3; i++) {
            accentVariations.push({transformations: [{type: 'lumOff', value: i * 8000}]});
        }

        for (var i = 0; i < allAccents.length; i++) {
            COLORSET.push({ name: 'Accent' + (i + 1), colors: [allAccents[i]], variations: accentVariations});
        }
    })();

    /**
     * an infoholder for the 48 chartstyles, a combination of colorpatterns, backgroundcolors and highlighting
     */
    var ChartStyleUtil = {

    };

    /**
     * return a list of different combinations of backgrounds and highlights
     */
    ChartStyleUtil.getStyleSet = function () {
        return STYLESET;
    };

    /**
     * return a list of different combinations of grayscales and accents
     */
    ChartStyleUtil.getColorSet = function () {
        return COLORSET;
    };

    /**
     * returns the color-transformations for desaturate the first color in the used pattern
     */
    ChartStyleUtil.getBackgroundTransformation = function () {
        return BACKGROUNDTRANSFORMATION;
    };

    function getCorrectLineColor(chart, styles) {
        var attrs = chart.getMergedAttributes();
        var bg = attrs.fill.color;
        if (bg.type === 'auto') {
            return Color.BLACK;
        }
        if (!_.isString(bg.fallbackValue)) {
            bg.fallbackValue = styles.getCssColor(bg);
        }
        if (Color.isDark(bg.fallbackValue)){
            return Color.WHITE;
        } else {
            return Color.BLACK;
        }
    }

    /**
     * sets all character-model-properties to the target,
     * target is a CanvasJS-data-model
     * @app {BaseApplication}
     * @character {object} character-model comes from operations
     * @target {object} target-object is a inner object of CanvasJS
     * @prefix {object} necessary if there are more than one character-holder on the target object
     *
     */
    ChartStyleUtil.handleCharacterProps = function (chart, character, target, prefix) {
        var app = chart.getApp();
        var view = app.getView();
        var styles = app.getModel().getDocumentStyles();

        var usePrefix;
        if (prefix && prefix.length) {
            usePrefix = prefix + 'Font';
        } else {
            usePrefix = 'font';
        }

        target[usePrefix + 'Color'] = styles.getCssColor(character.color, getCorrectLineColor(chart, styles));
        target[usePrefix + 'Family'] = styles.getCssFontFamily(character.fontName);
        target[usePrefix + 'Size'] = Utils.convertLength(view.getEffectiveZoom() * character.fontSize, 'pt', 'px', 0.01);
        target[usePrefix + 'Weight'] = character.bold ? 'bold' : 'normal';
        target[usePrefix + 'Style'] = character.italic ? 'italic' : 'normal';
    };

    /**
     * sets all lineshape-model-properties to the target,
     * target is a CanvasJS-data-model
     * @app {BaseApplication}
     * @character {object} lineshape-model comes from operations
     * @target {object} target-object is a inner object of CanvasJS
     * @prefix {object} necessary for more than one character-holder on the target object
     */
    ChartStyleUtil.handleLineProps = function (chart, line, target, prefix) {
        var app = chart.getApp();
        var view = app.getView();
        var styles = app.getModel().getDocumentStyles();
        var w = line.width || 1;
        var usePrefix = prefix;

        if (line.type === 'none') {
            target[usePrefix + 'Color'] = 'transparent';
        } else {
            target[usePrefix + 'Color'] = styles.getCssColor(line.color, getCorrectLineColor(chart, styles));
            target[usePrefix + 'Thickness'] = view.getEffectiveZoom() * w;
        }

    };

    /**
     * checks for assigned numberformat,
     * if there is only a code it looks in the clipboard-api for the code's numberformat
     *
     * if there is a 'm' in the numberformat, the code returns true because minutes or months are in Dates
     */
    ChartStyleUtil.isDate = function (value) {
        if (!value) {
            return;
        }
        var attrs = value.attrs;
        if (!attrs) {
            return;
        }

        var code = attrs.cell.numberFormat.code;
        if (!code || !code.length) {
            var cell = {
                attributes: attrs
            };
            code = Clipboard.getCalcNumberFormat(cell);
        }

        if (code && code.toLowerCase().indexOf('yy') >= 0) {
            return true;
        }

    };

    // exports ================================================================

    return ChartStyleUtil;

});
