/**
 * All content on this website (including text, images, source
 * code and any other original works), unless otherwise noted,
 * is licensed under a Creative Commons License.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) Open-Xchange Inc., 2006-2012
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/editframework/model/format/border',
    ['io.ox/office/tk/utils',
     'io.ox/office/editframework/model/format/color'
    ], function (Utils, Color) {

    'use strict';

    var // map operation line styles to CSS line styles (anything else maps to 'solid')
        CSS_LINE_STYLES = {
            none: 'none',
            double: 'double',
            triple: 'double',
            dashed: 'dashed',
            dashSmallGap: 'dashed',
            dotted: 'dotted',
            dotDash: 'dotted',
            dotDotDash: 'dotted',
            dashDotStroked: 'dotted'
        },

        // maximum line width for borders in GUI preview elements, in pixels
        MAX_PREVIEW_LINEWIDTH = 2,

        // minimum line width for borders with style 'double', in pixels
        MIN_DOUBLE_LINEWIDTH = 3;

    // static class Border ====================================================

    /**
     * Predefined border objects.
     */
    var Border = {
            NONE: { style: 'none' },
            SINGLE: { style: 'single', width: 17, space: 140, color: Color.AUTO }
        };

    /**
     * Returns the best CSS border style that can represent the passed line
     * style used in border attributes.
     *
     * @param {String} style
     *  The line style used in border attributes in operations.
     *
     * @returns {String}
     *  The best matching CSS border style.
     */
    Border.getCssBorderStyle = function (style) {
        return (style in CSS_LINE_STYLES) ? CSS_LINE_STYLES[style] : 'solid';
    };

    /**
     * Returns the effective CSS attributes for the passed border value.
     *
     * @param {Object} border
     *  The border object as used in operations.
     *
     * @param {Theme} theme
     *  The theme object used to map scheme color names to color values.
     *
     * @param {Object} [options]
     *  A map of options controlling the operation. Supports the following
     *  options:
     *  @param {Boolean} [options.preview=false]
     *      If set to true, the border will be rendered for a preview element
     *      in the GUI. The border width will be restricted to two pixels in
     *      that case.
     *  @param {Object} [options.autoColor]
     *      If specified, a replacement color to be used in case the border
     *      object contains the automatic color.
     *  @param {Boolean} [options.doubleWidth=false]
     *      If set to true, the border width will be increased by a factor of
     *      2 or even 4. This is caused by task 29381. IE 11 uses for border-collapse
     *      set to 'collapse' only half of the given width.
     *
     * @returns {Object}
     *  A map with CSS border attributes. Contains the property 'style' with
     *  the effective CSS border style as string; the property 'width' with the
     *  effective border width in pixels (as number, rounded to entire pixels);
     *  and the property 'color' with the effective CSS color (as string with
     *  leading hash sign).
     */
    Border.getCssBorderAttributes = function (border, theme, options) {

        var style = Utils.getStringOption(border, 'style', 'none'),
            width = Utils.getIntegerOption(border, 'width', 0),
            color = Utils.getObjectOption(border, 'color', Color.AUTO),
            preview = Utils.getBooleanOption(options, 'preview', false),
            autoColor = Utils.getObjectOption(options, 'autoColor'),
            doubleWidth = Utils.getBooleanOption(options, 'doubleWidth', false),
            factor = 1;

        // convert operation line styles to CSS styles
        style = Border.getCssBorderStyle(style);

        // convert 1/100mm to pixels (round to one digit)
        width = (style === 'none') ? 0 : Utils.convertHmmToLength(width, 'px', 0.1);

        // minimum width for borders with style 'double', maximum width for preview borders
        if ((style === 'double') && (width > 0)) {
            width = preview ? MIN_DOUBLE_LINEWIDTH : Math.max(width, MIN_DOUBLE_LINEWIDTH);
        } else if (preview) {
            width = Math.min(width, MAX_PREVIEW_LINEWIDTH);
        }

        // round to entire pixels
        if (width > 0) {
            if (doubleWidth) {
                // Fix for 29381, IE 11 handling for border-collapse.
                // The increased factor of 4 is necessary, because inner borders with width '1px'
                // (the minimum value) will be displayed as border with '2px' in IE 11.
                factor = width <= 1 ? 4 : 2;
                width *= factor;
            }
            width = Math.max(Math.round(width), 1);
        }

        // convert color object to CSS color
        color = Color.getCssColor(color, autoColor || 'line', theme);

        return { style: style, width: width, color: color };
    };

    /**
     * Converts the passed border attribute object to a CSS border value.
     *
     * @param {Object} border
     *  The border object as used in operations.
     *
     * @param {Theme} theme
     *  The theme object used to map scheme color names to color values.
     *
     * @param {Object} [options]
     *  A map of options controlling the operation. Supports the following
     *  options:
     *  @param {Boolean} [options.clearNone=false]
     *      If set to true, the return value for invisible borders will be the
     *      empty string instead of the keyword 'none'.
     *  @param {Boolean} [options.preview=false]
     *      If set to true, the border will be rendered for a preview element
     *      in the GUI. The border width will be restricted to two pixels in
     *      that case.
     *  @param {Object} [options.autoColor]
     *      If specified, a replacement color to be used in case the border
     *      object contains the automatic color.
     *
     * @returns {String}
     *  The CSS border value converted from the passed border object.
     */
    Border.getCssBorder = function (border, theme, options) {
        var visible = Border.isVisibleBorder(border),
            attributes = visible ? Border.getCssBorderAttributes(border, theme, options) : null;
        return visible ? (attributes.style + ' ' + attributes.width + 'px ' + attributes.color) :
            Utils.getBooleanOption(options, 'clearNone', false) ? '' : 'none';
    };

    /**
     * Returns whether the passed border object is a valid and visible border
     * style (line style different to 'none', and line width greater than 0).
     *
     * @param {Object} border
     *  The border object as used in operations.
     */
    Border.isVisibleBorder = function (border) {
        return _.isObject(border) && _.isString(border.style) && (border.style !== 'none') && _.isNumber(border.width) && (border.width > 0);
    };

    /**
     * Comparing two borders. These are equal if both borders have the same style, width and color
     *
     * @param {Border} border1
     * The first border to compare
     *
     * @param {Border} border2
     * The second border to compare
     *
     * @returns {Boolean}
     * Whether the two borders are equal.
     */
    Border.isEqual = function (border1, border2) {
        return (border1 &&
                border2 &&
                ((!('style' in border1) && (!('style' in border2))) || (('style' in border1) && ('style' in border2) && (border1.style === border2.style))) &&
                ((!('width' in border1) && (!('width' in border2))) || (('width' in border1) && ('width' in border2) && (border1.width === border2.width))) &&
                ((!('color' in border1) && (!('color' in border2))) || (('color' in border1) && ('color' in border2) && (Color.isEqual(border1.color, border2.color)))));
    };

    // exports ================================================================

    return Border;

});
