/**
 * All content on this website (including text, images, source
 * code and any other original works), unless otherwise noted,
 * is licensed under a Creative Commons License.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) Open-Xchange Inc., 2006-2012
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 * @author Ingo Schmidt-Rosbiegal <ingo.schmidt-rosbiegal@open-xchange.com>
 */

define('io.ox/office/spreadsheet/view/mixin/gridtrackingmixin',
    ['io.ox/office/tk/utils',
     'io.ox/office/drawinglayer/model/drawingutils',
     'io.ox/office/drawinglayer/view/drawingframe',
     'io.ox/office/spreadsheet/utils/sheetutils',
     'io.ox/office/spreadsheet/utils/paneutils'
    ], function (Utils, DrawingUtils, DrawingFrame, SheetUtils, PaneUtils) {

    'use strict';

    // class GridTrackingMixin ================================================

    /**
     * Mix-in class for the class GridPane that implements all kinds of mouse
     * and touch tracking, for example selection, or manipulation of drawing
     * objects.
     *
     * @constructor
     *
     * @param {SpreadsheetApplication} app
     *  The application that contains this grid pane.
     *
     * @param {String} panePos
     *  The position of this grid pane.
     *
     * @param {jQuery} layerRootNode
     *  The root node of all layer nodes contained in the grid pane.
     */
    function GridTrackingMixin(app, panePos, layerRootNode) {

        var // self reference
            self = this,

            // the spreadsheet model and view
            model = null,
            view = null,

            // the model and collections of the active sheet
            sheetModel = null,
            colCollection = null,
            rowCollection = null,
            drawingCollection = null,

            // the type of the current tracking, and the last finished tracking
            trackingType = null,
            lastTrackingType = null,

            // whether current tracking can be continued in read-only mode
            readOnlyTracking = false;

        // private methods ----------------------------------------------------

        /**
         * Handles a changed document edit mode. If switched to read-only mode,
         * cancels in-place cell edit mode, or a running tracking action to
         * move or resize drawing objects.
         */
        function editModeHandler(event, editMode) {
            // stop tracking if edit rights have been lost
            if (!editMode && !readOnlyTracking) {
                $.cancelTracking();
            }
        }

        /**
         * Initializes this grid pane after the active sheet has been changed.
         */
        function changeActiveSheetHandler(event, activeSheet, activeSheetModel) {
            sheetModel = activeSheetModel;
            colCollection = sheetModel.getColCollection();
            rowCollection = sheetModel.getRowCollection();
            drawingCollection = sheetModel.getDrawingCollection();
        }

        /**
         * Returns all important data for the passed tracking event.
         *
         * @param {jQuery.Event} event
         *  The tracking event.
         *
         * @returns {Object}
         *  A result map with the following properties:
         *  - {Number} left
         *      The horizontal absolute sheet position of the tracking point,
         *      in pixels.
         *  - {Number} top
         *      The vertical absolute sheet position of the tracking point, in
         *      pixels.
         *  - {Object} colEntry
         *      The column descriptor received from the column collection.
         *  - {Object} rowEntry
         *      The row descriptor received from the row collection.
         *  GridTrackingMixin.getTrackingPosition().
         *  - {Number[]} address
         *      The logical address of the cell covered by the tracking point.
         */
        function getTrackingData(event, options) {

            var // the absolute position of the layer root node
                layerOffset = layerRootNode.offset(),
                // the grid pane layer rectangle
                layerRectangle = self.getLayerRectangle(),
                // the result object
                result = {};

            // calculate sheet offset in pixels
            result.left = event.pageX - layerOffset.left + layerRectangle.left;
            result.top = event.pageY - layerOffset.top + layerRectangle.top;

            // add the column and row collection entry
            options = Utils.extendOptions(options, { pixel: true });
            result.colEntry = colCollection.getEntryByOffset(result.left, options);
            result.rowEntry = rowCollection.getEntryByOffset(result.top, options);

            // add the cell address
            result.address = [result.colEntry.index, result.rowEntry.index];

            return result;
        }

        /**
         * Returns the address of the anchor cell for a tracking cell range,
         * according to the passed tracking position.
         *
         * @param {jQuery.Event} event
         *  The initial 'tracking:start' event.
         *
         * @param {Object} origRange
         *  The logical address of the range to be tracked.
         *
         * @param {String} trackerPos
         *  The position identifier of the tracker node.
         *
         * @param {Boolean} resize
         *  Whether resize tracking (true), or border tracking (false) is
         *  active. For border tracking, the tracker position passed in the
         *  parameter 'trackerPos' must be 't', 'b', 'l', or 'r'.
         *
         * @returns {Number[]}
         *  The logical address of the anchor cell in the passed range.
         */
        function getAnchorCell(event, origRange, trackerPos, resize) {
            switch (trackerPos) {
            case 'l':
                return resize ? _.clone(origRange.end) : [origRange.start[0], getTrackingData(event).address[1]];
            case 'r':
                return resize ? _.clone(origRange.start) : [origRange.end[0], getTrackingData(event).address[1]];
            case 't':
                return resize ? _.clone(origRange.end) : [getTrackingData(event).address[0], origRange.start[1]];
            case 'b':
                return resize ? _.clone(origRange.start) : [getTrackingData(event).address[0], origRange.end[1]];
            case 'tl':
                return _.clone(origRange.end);
            case 'tr':
                return [origRange.start[0], origRange.end[1]];
            case 'bl':
                return [origRange.end[0], origRange.start[1]];
            case 'br':
                return _.clone(origRange.start);
            }
        }

        /**
         * Recalculates the address of the tracked range according to the
         * passed tracking position.
         *
         * @param {jQuery.Event} event
         *  The tracking event.
         *
         * @param {Object} origRange
         *  The original address of the range currently tracked.
         *
         * @param {Number[]} anchorCell
         *  The anchor cell inside the passed original range, as has been
         *  returned by the method GridTrackingMixin.getAnchorCell().
         *
         * @param {String} trackerPos
         *  The position identifier of the tracker node.
         *
         * @param {Boolean} resize
         *  Whether resize tracking (true), or border tracking (false) is
         *  active. For border tracking, the tracker position passed in the
         *  parameter 'trackerPos' must be 't', 'b', 'l', or 'r'.
         *
         * @returns {Object}
         *  The logical address of the current tracked range.
         */
        function recalcTrackedRange(event, origRange, anchorCell, trackerPos, resize) {

            var // current column/row entries
                trackingData = getTrackingData(event, { outerHidden: true }),
                // the cell address currently tracked
                address = trackingData.address,
                // the resulting range
                range = null;

            // adjusts column/row index according to the offset inside the covered cell
            function adjustTrackedBorder(columns) {

                var // the array index in cell addresses
                    index = columns ? 0 : 1,
                    // the maximum column/row index
                    maxIndex = columns ? model.getMaxCol() : model.getMaxRow(),
                    // the column/row descriptor
                    entryData = columns ? trackingData.colEntry : trackingData.rowEntry,
                    // whether the leading border is tracked
                    leadingBorder = resize ? (address[index] < anchorCell[index]) : (trackerPos === (columns ? 'l' : 't')),
                    // whether the trailing border is tracked
                    trailingBorder = resize ? (address[index] > anchorCell[index]) : (trackerPos === (columns ? 'r' : 'b')),
                    // the offset ratio in the tracked column/row
                    ratio = (entryData.size > 0) ? (entryData.relOffset / entryData.size) : Number.NaN;

                if (leadingBorder && (address[index] < maxIndex) && (ratio > 0.5)) {
                    // jump to next column/row if tracking leading border, and covering second half of the tracked column/row ...
                    address[index] += 1;
                } else if (trailingBorder && (address[index] > 0) && (ratio < 0.5)) {
                    // ... or to previous column/row if tracking trailing border, and covering first half of the tracked column/row
                    address[index] -= 1;
                }
            }

            // adjust column/row index according to the offset inside the tracked cell
            adjustTrackedBorder(false);
            adjustTrackedBorder(true);

            // build and return the new range according to the current position of the tracked cell
            if (resize) {
                // resize mode: build range between anchor cell and tracked cell
                range = SheetUtils.getAdjustedRange({ start: anchorCell, end: address });
                if ((trackerPos === 't') || (trackerPos === 'b')) {
                    // resize vertically: restore row indexes to original range
                    range.start[0] = origRange.start[0];
                    range.end[0] = origRange.end[0];
                } else if ((trackerPos === 'l') || (trackerPos === 'r')) {
                    // resize horizontally: restore row indexes to original range
                    range.start[1] = origRange.start[1];
                    range.end[1] = origRange.end[1];
                }
            } else {
                // border mode: move entire range according to distance between tracked cell and anchor cell
                range = model.getBoundedMovedRange(origRange, address[0] - anchorCell[0], address[1] - anchorCell[1]);
            }

            return range;
        }

        // cell selection (mouse) ---------------------------------------------

        /**
         * Handles all tracking events for cell selection originating from
         * mouse events.
         */
        var mouseSelectionTrackingHandler = (function () {

            var // current address (prevent updates while tracking over the same cell)
                currentAddress = null;

            // updates the current address according to the passed tracking event
            function updateCurrentAddress(event) {
                var address = getTrackingData(event, { outerHidden: true }).address;
                if (!_.isEqual(address, currentAddress)) {
                    currentAddress = address;
                    self.trigger('select:move', currentAddress);
                }
            }

            // return the actual mouseSelectionTrackingHandler() method
            return function (event) {
                switch (event.type) {
                case 'tracking:start':
                    currentAddress = getTrackingData(event).address;
                    self.trigger('select:start', currentAddress, PaneUtils.getSelectionMode(event));
                    self.scrollToCell(currentAddress);
                    break;
                case 'tracking:move':
                    // prevents that Chrome modifies the text area selection while moving mouse
                    event.preventDefault();
                    updateCurrentAddress(event);
                    break;
                case 'tracking:scroll':
                    self.scrollRelative(event.scrollX, event.scrollY);
                    updateCurrentAddress(event);
                    break;
                case 'tracking:end':
                    self.trigger('select:end', currentAddress, true);
                    self.scrollToCell(currentAddress);
                    break;
                case 'tracking:cancel':
                    self.trigger('select:end', currentAddress, false);
                    self.scrollToCell(currentAddress);
                    break;
                }
            };
        }()); // end of mouseSelectionTrackingHandler() local scope

        // cell selection (touch) ---------------------------------------------

        /**
         * Handles all tracking events for cell selection originating from
         * touch events.
         */
        var touchSelectionTrackingHandler = (function () {

            var // cell address where tracking has started
                startAddress = null,
                // whether the touch point has been moved
                moved = false;

            function finalizeTracking(event) {

                var // current selection
                    selection = view.getSelection(),
                    // array index of the range that contains a tapped cell
                    rangeIndex = 0;

                // do nothing, if the touch point has been moved
                if (moved) { return; }

                if (event.duration <= 750) {
                    // short-tap: select a single cell, or start in-place edit mode
                    if (!view.hasDrawingSelection() && _.isEqual(selection.activeCell, startAddress)) {
                        view.enterCellEditMode();
                    } else {
                        self.trigger('select:start', startAddress, 'select');
                        self.trigger('select:end', startAddress, true);
                        self.scrollToCell(startAddress);
                    }
                } else {
                    // long-tap: change active cell, if tapped inside selection
                    rangeIndex = Utils.findFirstIndex(selection.ranges, function (range) {
                        return SheetUtils.rangeContainsCell(range, startAddress);
                    });
                    if (rangeIndex >= 0) {
                        self.trigger('select:active', rangeIndex, startAddress);
                        self.scrollToCell(startAddress);
                    }
                }
            }

            // return the actual touchSelectionTrackingHandler() method
            return function (event) {
                switch (event.type) {
                case 'tracking:start':
                    startAddress = getTrackingData(event).address;
                    moved = false;
                    // prevent following mouse events after short-tap on iPad
                    event.preventDefault();
                    break;
                case 'tracking:move':
                    self.scrollRelative(-event.moveX, -event.moveY);
                    if (Math.abs(event.offsetX) + Math.abs(event.offsetY) >= 6) { moved = true; }
                    // prevent native scrolling
                    event.preventDefault();
                    break;
                case 'tracking:end':
                    finalizeTracking(event);
                    break;
                case 'tracking:cancel':
                    break;
                }
            };
        }()); // end of touchSelectionTrackingHandler() local scope

        // resize selection range (touch) -------------------------------------

        /**
         * Handles tracking events to resize selected ranges on touch devices.
         */
        var resizeSelectionTrackingHandler = (function () {

            var // the position identifier of the tracked node
                trackerPos = null,
                // the array index of the tracked range
                rangeIndex = 0,
                // the original address of the tracked range
                origRange = null,
                // the anchor cell in the original range
                anchorCell = null,
                // the current address of the tracked range
                currRange = null;

            // initializes the highlighted range according to the passed tracking event
            function initializeTracking(event) {

                var // the handle node currently tracked
                    trackerNode = $(event.target),
                    // the root node of the range that will be tracked
                    rangeNode = trackerNode.closest('.range');

                // extract range address
                rangeIndex = Utils.getElementAttributeAsInteger(rangeNode, 'data-index', -1);
                origRange = view.getSelectedRanges()[rangeIndex];

                // initialize additional data needed while tracking
                trackerPos = trackerNode.attr('data-pos');
                currRange = _.copy(origRange, true);
                anchorCell = getAnchorCell(event, origRange, trackerPos, true);
            }

            // updates the highlighted range according to the passed tracking event
            function updateTracking(event) {

                var // the new range address
                    newRange = recalcTrackedRange(event, origRange, anchorCell, trackerPos, true);

                // update the tracked range in the respective token array, this will
                // trigger change events which will cause refreshing all grid panes
                if (!_.isEqual(currRange, newRange)) {
                    currRange = newRange;
                    self.trigger('select:range', rangeIndex, currRange);
                }
            }

            // return the actual resizeSelectionTrackingHandler() method
            return function (event) {
                switch (event.type) {
                case 'tracking:start':
                    initializeTracking(event);
                    event.preventDefault(); // prevent native scrolling
                    break;
                case 'tracking:move':
                    updateTracking(event);
                    event.preventDefault(); // prevent native scrolling
                    break;
                case 'tracking:scroll':
                    self.scrollRelative(event.scrollX, event.scrollY);
                    updateTracking(event);
                    break;
                }
            };
        }()); // end of resizeSelectionTrackingHandler() local scope

        // move/resize highlighted ranges -------------------------------------

        /**
         * Handles all tracking events for highlighted ranges.
         */
        var highlightTrackingHandler = (function () {

            var // the position identifier of the tracked node (border or corner handle)
                trackerPos = null,
                // whether resize tracking is active
                resize = false,
                // the original address of the tracked range
                origRange = null,
                // the anchor cell in the original range
                anchorCell = null,
                // the current address of the tracked range
                currRange = null;

            // initializes the highlighted range according to the passed tracking event
            function initializeTracking(event) {

                var // the border/handle node currently tracked
                    trackerNode = $(event.target),
                    // the root node of the highlighted range that will be tracked
                    rangeNode = trackerNode.closest('.range'),
                    // the highlighted ranges
                    highlightRanges = sheetModel.getViewAttribute('highlightRanges'),
                    // the array index of the tracked range
                    highlightIndex = Utils.getElementAttributeAsInteger(rangeNode, 'data-index', -1);

                // extract original range from the highlight ranges
                origRange = _.isArray(highlightRanges) ? highlightRanges[highlightIndex] : null;
                if (!origRange) {
                    Utils.warn('GridTrackingMixin.initializeTracking(): cannot find tracking range');
                    $.cancelTracking();
                    return;
                }

                // initialize additional data needed while tracking
                trackerPos = trackerNode.attr('data-pos');
                resize = trackerNode.parent().hasClass('resizers');
                currRange = { start: origRange.start, end: origRange.end };
                anchorCell = getAnchorCell(event, origRange, trackerPos, resize);

                // render tracked range with special highlighting
                sheetModel.setViewAttribute('highlightIndex', highlightIndex);
            }

            // updates the highlighted range according to the passed tracking event
            function updateTracking(event) {

                var // the new range address
                    newRange = recalcTrackedRange(event, origRange, anchorCell, trackerPos, resize);

                // update the tracked range in the respective token array, this will
                // trigger change events which will cause refreshing all grid panes
                if (!_.isEqual(currRange, newRange)) {
                    currRange = newRange;
                    view.modifyReferenceToken(origRange.id, currRange);
                }
            }

            // finalizes the highlighted range
            function finalizeTracking() {
                sheetModel.setViewAttribute('highlightIndex', null);
            }

            // return the actual highlightTrackingHandler() method
            return function (event) {
                switch (event.type) {
                case 'tracking:start':
                    initializeTracking(event);
                    event.preventDefault();
                    break;
                case 'tracking:move':
                    updateTracking(event);
                    // prevent Chrome from modifying the text area selection while moving mouse
                    event.preventDefault();
                    break;
                case 'tracking:scroll':
                    self.scrollRelative(event.scrollX, event.scrollY);
                    updateTracking(event);
                    break;
                case 'tracking:end':
                    finalizeTracking(event);
                    break;
                case 'tracking:cancel':
                    finalizeTracking(event);
                    view.modifyReferenceToken(origRange.id, origRange);
                    break;
                }
            };
        }()); // end of highlightTrackingHandler() local scope

        // auto fill ----------------------------------------------------------

        /**
         * Handles all tracking events for the auto-fill handle.
         */
        var autoFillTrackingHandler = (function () {

            var // the initial selection range
                origRange = null,
                // the sheet rectangle covered by the initial selection range
                origRectangle = null,
                // start position of tracking (bottom-right corner of selection range)
                startPos = null,
                // whether to restrict tracking direction (column/row ranges)
                fixedColumns = null,
                // the current value for the 'autoFillData' attribute
                autoFillData = null;

            // initializes the auto fill handling according to the passed tracking event
            function initializeTracking() {

                // store address and position of the original selection range
                origRange = view.getSelectedRanges()[0];
                origRectangle = sheetModel.getRangeRectangle(origRange);
                startPos = { left: origRectangle.left + origRectangle.width, top: origRectangle.top + origRectangle.height };

                // special behavior for column/row ranges
                fixedColumns = model.isColRange(origRange) ? true : model.isRowRange(origRange) ? false : null;

                // render selected range with modified auto-fill style
                autoFillData = { border: 'bottom', count: 0 };
                sheetModel.setViewAttribute('autoFillData', autoFillData);
            }

            // updates the auto fill area according to the passed tracking event
            function updateTracking(event) {

                var // current tracking position
                    data = getTrackingData(event, { outerHidden: true }),
                    // whether to delete parts of the selected range
                    deleteMode = DrawingUtils.rectangleContainsPosition(origRectangle, data),
                    // the adjusted tracking position to determine the direction
                    left = (deleteMode || (data.left >= startPos.left)) ? data.left : Math.min(data.left + origRectangle.width, startPos.left),
                    top = (deleteMode || (data.top >= startPos.top)) ? data.top : Math.min(data.top + origRectangle.height, startPos.top),
                    // whether to expand horizontally or vertically
                    columns = _.isBoolean(fixedColumns) ? fixedColumns : (Math.abs(top - startPos.top) < Math.abs(left - startPos.left)),
                    // whether to expand the leading or trailing border
                    leading = columns ? (left < startPos.left) : (top < startPos.top),
                    // array index in cell addresses
                    addrIndex = columns ? 0 : 1,
                    // the tracker position to calculate the target rectangle
                    trackerPos = columns ? (leading ? 'l' : 'r') : (leading ? 't' : 'b'),
                    // simulate tracking at a specific border of the selected range
                    range = recalcTrackedRange(event, origRange, leading ? origRange.start : origRange.end, trackerPos, false);

                // calculate the auto-fill data from target range
                if (deleteMode) {
                    // delete mode: always delete from trailing border
                    autoFillData.border = columns ? 'right' : 'bottom';
                    autoFillData.count = range.start[addrIndex] - origRange.end[addrIndex] - 1;
                } else if (leading) {
                    // expand leading range border
                    autoFillData.border = columns ? 'left' : 'top';
                    autoFillData.count = origRange.start[addrIndex] - range.start[addrIndex];
                } else {
                    // expand trailing range border
                    autoFillData.border = columns ? 'right' : 'bottom';
                    autoFillData.count = range.end[addrIndex] - origRange.end[addrIndex];
                }
                sheetModel.setViewAttribute('autoFillData', autoFillData);

                // set correct mouse pointer according to direction
                event.cursor = (_.isBoolean(fixedColumns) || (autoFillData.count === 0)) ? null : columns ? 'ew-resize' : 'ns-resize';
            }

            // finalizes the selection tracking
            function finalizeTracking() {
                sheetModel.setViewAttribute('autoFillData', null);
            }

            // return the actual autoFillTrackingHandler() method
            return function (event) {

                switch (event.type) {
                case 'tracking:start':
                    initializeTracking(event);
                    break;
                case 'tracking:move':
                    updateTracking(event);
                    break;
                case 'tracking:scroll':
                    self.scrollRelative(event.scrollX, event.scrollY);
                    updateTracking(event);
                    break;
                case 'tracking:end':
                    finalizeTracking(event);
                    view.autoFill(autoFillData.border, autoFillData.count);
                    break;
                case 'tracking:cancel':
                    finalizeTracking(event);
                    break;
                }
            };
        }()); // end of autoFillTrackingHandler() local scope

        // select/move/resize drawings ----------------------------------------

        /**
         * Handles all tracking events for drawing frames.
         */
        var drawingTrackingHandler = (function () {

            var // the current drawing frame being tracked
                drawingFrame = null,
                // the logical position of the tracked drawing frame
                drawingPos = null,
                // the container element used to visualize the movement and resizing
                moveBox = null,
                // the position change in the sheet (in px)
                shiftX = 0, shiftY = 0,
                // the start position of the event in the sheet
                startX = 0, startY = 0,
                // the current width and height of the move box
                currentWidth = null, currentHeight = null,
                // the current left and top position of the move box
                currentLeft = null, currentTop = null,
                // whether the mouse was really moved
                mouseMoved = false,
                // whether the drawing shall be moved or resized
                doResize = false,
                // whether resizing is available in horizontal/vertical direction
                useX = false, useY = false, leftResize = false, topResize = false,
                // the maximum allowed values for move and resize
                maxLeft = 0, maxTop = 0, maxRight = 0, maxBottom = 0,
                // the width-to-height ratio of the drawing
                ratio = 1,
                // the width and height of the 'active' drawing before resizing it, in px
                activeStartWidth = 0, activeStartHeight = 0,
                // whether more than one drawing is selected
                isMultiSelection = false,
                // multi selection: container for all positions of selected drawings
                allDrawingPositions = [],
                // multi selection: container for all drawing frames of selected drawings
                allDrawingFrames = [],
                // multi selection: container for all drawing move boxes of selected drawings
                allDrawingMoveBoxes = [],
                // multi selection: container for all drawing geometries
                allStartWidths = [], allStartHeights = [];

            // initializes the drawing according to the passed tracking event
            function initializeTracking(event) {

                var // the grid pane layer rectangle
                    layerRectangle = self.getLayerRectangle(),
                    // whether to toggle the selection of the clicked drawing frame
                    toggle = event.shiftKey || event.ctrlKey || event.metaKey,
                    // whether the clicked drawing frame is already selected
                    selected = false,
                    // taking care of the distance to the border and of the geometry of each drawing during resize
                    widthRatio = 1, heightRatio = 1,
                    // the resize handle position
                    pos = null,
                    // the current selection
                    selection = null,
                    // the selected drawing objects
                    drawings = null,
                    // the drawing attributes
                    oneDrawingAttributes = null, drawingAttributes = null,
                    // the width and height of the 'active' drawing before resizing it, in hmm
                    activeStartWidthHmm = 0, activeStartHeightHmm = 0;

                // ensure that a drawing frame has been hit
                if (!(drawingFrame = Utils.findFarthest(layerRootNode, event.target, DrawingFrame.NODE_SELECTOR))) {
                    $.cancelTracking();
                    return;
                }

                // convert to single jQuery object
                drawingFrame = $(drawingFrame).first();

                // do not modify the selection, if clicked on a selected drawing without modifier key
                selected = DrawingFrame.isSelected(drawingFrame);
                if (toggle || !selected) {
                    drawingPos = [$(drawingFrame).index()];
                    view.selectDrawing(drawingPos, { toggle: toggle });
                    view.scrollToSelectedDrawing(drawingPos);
                }

                // check edit mode and sheet protection after selecting the drawing
                if (!model.getEditMode() || view.getActiveSheetModel().isLocked()) {
                    $.cancelTracking();
                    return;
                }

                // finding moveBox, if not yet found
                if (!moveBox) {
                    moveBox = drawingFrame.find('.tracker');
                    if (moveBox.length === 0) {
                        moveBox = null;
                    }
                }

                // cancel tracking, if moveBox cannot be found
                if (!moveBox) {
                    $.cancelTracking();
                    return;
                }

                // receiving the current selection
                selection = view.getSelection();

                isMultiSelection = (selection.drawings.length > 1);

                // Handling multi selection as special case for performance reasons
                if (isMultiSelection) {
                    drawings = view.getSelectedDrawings();

                    // resetting all drawing collectors
                    allDrawingFrames = [];
                    allDrawingMoveBoxes = [];
                    allDrawingPositions = [];
                    allStartWidths = [];
                    allStartHeights = [];

                    _(drawings).each(function (position) {

                        var localDrawingFrame = self.getDrawingFrame(drawingCollection.findModel(position).getUid()).first(),
                            localMoveBox = localDrawingFrame.find('.tracker');

                        // collecting drawings for multi selection handling
                        if (localDrawingFrame && localMoveBox) {
                            allDrawingFrames.push(localDrawingFrame);
                            allDrawingMoveBoxes.push(localMoveBox);
                            allDrawingPositions.push(_.clone(position));
                        }
                    });
                }

                // setting correct size of moveBox (required after undo of drawing attributes on selected drawing
                if (isMultiSelection) {
                    _(allDrawingMoveBoxes).each(function (localMoveBox, index) {
                        localMoveBox.css({
                            left: 0,
                            top: 0,
                            width: allDrawingFrames[index].width(),
                            height: allDrawingFrames[index].height()
                        });
                    });
                } else {
                    moveBox.css({
                        left: 0,
                        top: 0,
                        width: drawingFrame.width(),
                        height: drawingFrame.height()
                    });
                }

                // check, whether the drawing shall be moved or resized
                pos = DrawingFrame.getResizerHandleType(event.target);
                if (_.isString(pos)) {
                    doResize = true;

                    // collecting information about the handle node
                    useX = /[lr]/.test(pos);
                    useY = /[tb]/.test(pos);
                    leftResize = /[l]/.test(pos);
                    topResize = /[t]/.test(pos);

                    if (isMultiSelection) {
                        _(allDrawingFrames).each(function (oneDrawingFrame) {
                            allStartWidths.push(oneDrawingFrame.width());
                            allStartHeights.push(oneDrawingFrame.height());
                        });
                    }

                    // geometry of the 'active' drawing
                    activeStartWidth = drawingFrame.width();
                    activeStartHeight = drawingFrame.height();
                    // receiving explicit attributes of the drawing
                    drawingAttributes = DrawingFrame.getModel(drawingFrame).getExplicitAttributes().drawing;
                    activeStartWidthHmm = drawingAttributes.width;
                    activeStartHeightHmm = drawingAttributes.height;
                    // using hmm to calculate ratio instead of px because of increased precision
                    ratio = Utils.round(activeStartWidthHmm / activeStartHeightHmm, 0.01);

                } else {
                    doResize = false;
                    useX = false;
                    useY = false;
                }

                startX = event.pageX - layerRootNode.offset().left + layerRectangle.left;
                startY = event.pageY - layerRootNode.offset().top + layerRectangle.top;

                // the maximum allowed change to the left, top, right and bottom
                if (isMultiSelection) {
                    _(allDrawingFrames).each(function (oneDrawingFrame, index) {
                        if (doResize) {
                            // receiving explicit attributes of the drawing
                            oneDrawingAttributes = DrawingFrame.getModel(oneDrawingFrame).getExplicitAttributes().drawing;
                            // taking care of the distance to the border and of the geometry of each drawing
                            widthRatio = Utils.round(activeStartWidthHmm / oneDrawingAttributes.width, 0.01);
                            heightRatio = Utils.round(activeStartHeightHmm / oneDrawingAttributes.height, 0.01);
                        } else {  // no change of image size during move of drawing
                            widthRatio = 1;
                            heightRatio = 1;
                        }
                        if (index === 0) {
                            maxLeft = Math.round((oneDrawingFrame.offset().left - layerRootNode.offset().left + layerRectangle.left) * widthRatio);
                            maxTop = Math.round((oneDrawingFrame.offset().top - layerRootNode.offset().top + layerRectangle.top) * heightRatio);
                            maxRight = Math.round((colCollection.getTotalSize() - layerRectangle.left - oneDrawingFrame.offset().left + layerRootNode.offset().left - oneDrawingFrame.width()) * widthRatio);
                            maxBottom = Math.round((rowCollection.getTotalSize() - layerRectangle.top - oneDrawingFrame.offset().top + layerRootNode.offset().top - oneDrawingFrame.height()) * heightRatio);
                        } else {
                            maxLeft = Math.min(maxLeft, Math.round((oneDrawingFrame.offset().left - layerRootNode.offset().left + layerRectangle.left) * widthRatio));
                            maxTop = Math.min(maxTop, Math.round((oneDrawingFrame.offset().top - layerRootNode.offset().top + layerRectangle.top) * heightRatio));
                            maxRight = Math.min(maxRight, Math.round((colCollection.getTotalSize() - layerRectangle.left - oneDrawingFrame.offset().left + layerRootNode.offset().left - oneDrawingFrame.width()) * widthRatio));
                            maxBottom = Math.min(maxBottom, Math.round((rowCollection.getTotalSize() - layerRectangle.top - oneDrawingFrame.offset().top + layerRootNode.offset().top - oneDrawingFrame.height()) * heightRatio));
                        }
                    });
                } else {
                    maxLeft = Math.round(drawingFrame.offset().left - layerRootNode.offset().left + layerRectangle.left);
                    maxTop = Math.round(drawingFrame.offset().top - layerRootNode.offset().top + layerRectangle.top);
                    maxRight = Math.round(colCollection.getTotalSize() - layerRectangle.left - drawingFrame.offset().left + layerRootNode.offset().left - drawingFrame.width());
                    maxBottom = Math.round(rowCollection.getTotalSize() - layerRectangle.top - drawingFrame.offset().top + layerRootNode.offset().top - drawingFrame.height());
                }

                // keeping the width-to-height ratio constant in resize mode
                if (doResize && useX && useY) {
                    if (leftResize && topResize) {
                        if ((maxLeft / maxTop) > ratio) { maxLeft = Math.round(maxTop * ratio); }
                        else { maxTop = Math.round(maxLeft / ratio); }
                    } else if (leftResize && !topResize) {
                        if ((maxLeft / maxBottom) > ratio) { maxLeft = Math.round(maxBottom * ratio); }
                        else { maxBottom = Math.round(maxLeft / ratio); }
                    } else if (!leftResize && topResize) {
                        if ((maxRight / maxTop) > ratio) { maxRight = Math.round(maxTop * ratio); }
                        else { maxTop = Math.round(maxRight / ratio); }
                    } else if (!leftResize && !topResize) {
                        if ((maxRight / maxBottom) > ratio) { maxRight = Math.round(maxBottom * ratio); }
                        else { maxBottom = Math.round(maxRight / ratio); }
                    }
                }
            }

            // calculating the attributes for a resized drawing during the resize action
            function getCurrentDrawingResizeAttributes(localMoveBox, startWidth, startHeight, options) {

                var // the horizontal and vertical scaling factors of the drawing
                    scaleWidth = 1, scaleHeight = 1,
                    // attributes for setting the position of the movebox
                    attrs = {},
                    // local variables for the shift in horizontal and vertical direction
                    localShiftX = shiftX, localShiftY = shiftY,
                    // whether an object with attributes shall be returned
                    getAttributes = Utils.getBooleanOption(options, 'getAttributes', true);

                currentWidth = null;
                currentHeight = null;
                currentLeft = null;
                currentTop = null;

                if (useX && !useY) {
                    localShiftY = 0;
                    currentHeight = startHeight;

                    if (isMultiSelection) {
                        // Modifying the shift for each drawing
                        localShiftX = Math.round(shiftX * startWidth / activeStartWidth);
                    }

                    if (leftResize) {
                        currentLeft = localShiftX;
                        currentWidth = startWidth - localShiftX;
                    } else {
                        currentWidth = startWidth + localShiftX;
                    }
                } else if (useY && !useX) {
                    localShiftX = 0;
                    currentWidth = startWidth;

                    if (isMultiSelection) {
                        // Modifying the shift for each drawing
                        localShiftY = Math.round(shiftY * startHeight / activeStartHeight);
                    }

                    if (topResize) {
                        currentTop = localShiftY;
                        currentHeight = startHeight - localShiftY;
                    } else {
                        currentHeight = startHeight + localShiftY;
                    }
                } else if (useX && useY) {

                    if (isMultiSelection) {
                        // Modifying the shift for each drawing
                        // -> localShiftX and localShiftY are yet adapted to the 'active' drawing
                        localShiftX = Math.round(shiftX * startWidth / activeStartWidth);
                        localShiftY = Math.round(shiftY * startHeight / activeStartHeight);
                    }

                    currentWidth = leftResize ? (startWidth - localShiftX) : (startWidth + localShiftX);
                    currentHeight = topResize ? (startHeight - localShiftY) : (startHeight + localShiftY);

                    // use the same scaling factor for vertical and horizontal resizing, if both are enabled -> the larger factor wins
                    scaleWidth = Utils.round(currentWidth / startWidth, 0.01);
                    scaleHeight = Utils.round(currentHeight / startHeight, 0.01);

                    if (scaleWidth > scaleHeight) {
                        currentHeight = Math.round(scaleWidth * startHeight);
                        localShiftY = topResize ? (startHeight - currentHeight) : (currentHeight - startHeight);
                    } else {
                        currentWidth = Math.round(scaleHeight * startWidth);
                        localShiftX = leftResize ? (startWidth - currentWidth) : (currentWidth - startWidth);
                    }

                    if (leftResize) { currentLeft = localShiftX; }
                    if (topResize) { currentTop = localShiftY; }
                }

                // Modifying the global values of shiftX, if this is not a multi selection. Otherwise shiftX and shiftY
                // need to be calculated for each drawing individually
                if (! isMultiSelection) {
                    shiftX = localShiftX;
                    shiftY = localShiftY;
                }

                // defining resize attributes
                if (_.isNumber(currentLeft)) { attrs.left = currentLeft; }
                if (_.isNumber(currentTop)) { attrs.top = currentTop; }
                if (_.isNumber(currentWidth)) { attrs.width = currentWidth; }
                if (_.isNumber(currentHeight)) { attrs.height = currentHeight; }

                return getAttributes ? attrs : { shiftX: localShiftX, shiftY: localShiftY };
            }

            // updates the position of the drawing frame according to the passed tracking event
            function updateDrawingPosition(event) {

                var // the grid pane layer rectangle
                    layerRectangle = self.getLayerRectangle(),
                    // the current position of the event in the sheet
                    currentX = 0, currentY = 0,
                    // the minimal size for a drawing side in pixel
                    minimumSize = 8;

                if (moveBox) {

                    currentX = event.pageX - layerRootNode.offset().left + layerRectangle.left;
                    currentY = event.pageY - layerRootNode.offset().top + layerRectangle.top;

                    shiftX = currentX - startX;
                    shiftY = currentY - startY;

                    // restricting values of shiftX and shiftY
                    // -> comparing shifts with 'maxTop', 'maxRight', 'maxBottom', 'maxLeft'
                    if ((leftResize || !doResize) && (-shiftX >= maxLeft)) {
                        shiftX = - maxLeft;
                        if (useX && useY) { shiftY = Math.round(shiftX / ratio); }
                    }

                    if ((topResize || !doResize) && (-shiftY >= maxTop)) {
                        shiftY = - maxTop;
                        if (useX && useY) { shiftX = Math.round(shiftY * ratio); }
                    }

                    if (((useX && !leftResize) || !doResize) && (shiftX >= maxRight)) {
                        shiftX = maxRight;
                        if (useX && useY) { shiftY = Math.round(shiftX / ratio); }
                    }

                    if (((useY && !topResize) || !doResize) && (shiftY >= maxBottom)) {
                        shiftY = maxBottom;
                        if (useX && useY) { shiftX = Math.round(shiftY * ratio); }
                    }

                    // ignoring specific shifts  for stretching of drawings
                    if (doResize && useX && !useY) { shiftY = 0; }
                    if (doResize && !useX && useY) { shiftX = 0; }

                    // avoid also decreasing the drawing too much
                    // -> comparing shifts with 'activeStartWidth' and 'activeStartHeight'
                    if (doResize) {
                        if (useX && useY) {
                            if ((leftResize) && (shiftX >= (activeStartWidth - minimumSize))) {
                                if ((topResize) && (shiftY >= (activeStartHeight - minimumSize))) {
                                    shiftX = activeStartWidth - minimumSize;
                                    shiftY = Math.round(shiftX / ratio);
                                } else if ((!topResize) && (-shiftY >= (activeStartHeight - minimumSize))) {
                                    shiftX = activeStartWidth - minimumSize;
                                    shiftY = - Math.round(shiftX / ratio);
                                }
                            } else if ((!leftResize) && (-shiftX >= (activeStartWidth - minimumSize))) {
                                if ((topResize) && (shiftY >= (activeStartHeight - minimumSize))) {
                                    shiftX = - activeStartWidth + minimumSize;
                                    shiftY = - Math.round(shiftX / ratio);
                                } else if ((!topResize) && (-shiftY >= (activeStartHeight - minimumSize))) {
                                    shiftX = - activeStartWidth + minimumSize;
                                    shiftY = Math.round(shiftX / ratio);
                                }
                            }
                        } else {
                            if ((leftResize) && (shiftX >= (activeStartWidth - minimumSize))) {
                                shiftX = activeStartWidth - minimumSize;
                            } else if ((!leftResize) && (useX) && (-shiftX >= (activeStartWidth - minimumSize))) {
                                shiftX = - activeStartWidth + minimumSize;
                            }
                            if ((topResize) && (shiftY >= (activeStartHeight - minimumSize))) {
                                shiftY = activeStartHeight - minimumSize;
                            } else if ((!topResize) && (useY) && (-shiftY >= (activeStartHeight - minimumSize))) {
                                shiftY = - activeStartHeight + minimumSize;
                            }
                        }
                    }

                    // only move the moveBox, if the mouse was really moved
                    mouseMoved = ((shiftX !== 0) || (shiftY !== 0));

                    if (mouseMoved) {

                        // make move box visible when tracking position has moved

                        if (isMultiSelection) {
                            _(allDrawingFrames).each(function (drawingFrame) {
                                DrawingFrame.toggleTracking(drawingFrame, true);
                            });
                        } else {
                            DrawingFrame.toggleTracking(drawingFrame, true);
                        }

                        if ((_.isNumber(shiftX)) && (_.isNumber(shiftY)) && (shiftX !== 0) || (shiftY !== 0)) {
                            if (! doResize) {
                                // moving the drawing(s)
                                if (isMultiSelection) {
                                    _(allDrawingMoveBoxes).each(function (localMoveBox) {
                                        localMoveBox.css({ left: shiftX, top: shiftY });
                                    });
                                } else {
                                    moveBox.css({ left: shiftX, top: shiftY });
                                }
                            } else {
                                // resizing the drawing(s)
                                if (isMultiSelection) {
                                    _(allDrawingMoveBoxes).each(function (localMoveBox, index) {
                                        localMoveBox.css(getCurrentDrawingResizeAttributes(localMoveBox, allStartWidths[index], allStartHeights[index]));
                                    });
                                } else {
                                    moveBox.css(getCurrentDrawingResizeAttributes(moveBox, activeStartWidth, activeStartHeight));
                                }
                            }
                        }
                    }
                }
            }

            // sets the drawing attributes to the drawing using view.setDrawingAttributes
            function setDrawingAttributes(localMoveBox, localDrawingFrame, index) {

                var // the grid pane layer rectangle
                    layerRectangle = self.getLayerRectangle(),
                    // the new drawing corner positions in 'px'
                    newTop = 0, newLeft = 0, newWidth = 0, newHeight = 0,
                    // the shift in each direction specific for each drawing
                    localShiftX = shiftX, localShiftY = shiftY,
                    // the calculated drawing attributes
                    attributes = {},
                    // a helper object to contain drawing specific shifts in multi drawing selections
                    shiftObject = {};

                if (! doResize) {
                    // moving the drawings
                    newLeft = localDrawingFrame.position().left + localShiftX + layerRectangle.left;
                    newTop = localDrawingFrame.position().top + localShiftY + layerRectangle.top;
                    newWidth = localDrawingFrame.width();
                    newHeight = localDrawingFrame.height();
                } else {
                    // resizing the drawings
                    if (isMultiSelection) {
                        // Modifying the shift for each drawing using function from 'updateDrawingPosition'
                        shiftObject = getCurrentDrawingResizeAttributes(localMoveBox, allStartWidths[index], allStartHeights[index], { getAttributes: false });
                        localShiftX = shiftObject.shiftX;
                        localShiftY = shiftObject.shiftY;
                    }

                    newLeft = leftResize ? localDrawingFrame.position().left + localShiftX : localDrawingFrame.position().left;
                    newTop = topResize ? localDrawingFrame.position().top + localShiftY : localDrawingFrame.position().top;
                    newLeft += layerRectangle.left;
                    newTop += layerRectangle.top;
                    newWidth = localMoveBox.width();
                    newHeight = localMoveBox.height();
                }

                // modifying attributes
                attributes = drawingCollection.getAttributesForRectangle({ left: newLeft, top: newTop, width: newWidth, height: newHeight});

                // setting moveBox back into the drawing frame
                localMoveBox.css({ left: 0, top: 0 });

                // generating drawing attributes operation(s)
                if (isMultiSelection) {
                    view.setDrawingAttributes(attributes, [allDrawingPositions[index]]);
                } else {
                    view.setDrawingAttributes(attributes);
                }
            }

            // finalizes the drawing tracking
            function finalizeDrawingTracking(event, applyPosition) {

                // generate an operation only, if the drawing was really moved
                if (mouseMoved && moveBox && applyPosition) {
                    if (isMultiSelection) {
                        _(allDrawingMoveBoxes).each(function (oneMoveBox, index) {
                            // index can be used for drawingFrames and drawingPositions, too
                            setDrawingAttributes(oneMoveBox, allDrawingFrames[index], index);
                        });
                    } else {
                        setDrawingAttributes(moveBox, drawingFrame);
                    }
                }

                // Clarify: setting moveBox to null is important for further selections of drawing
                if (isMultiSelection) {
                    _(allDrawingFrames).each(function (drawingFrame) {
                        // make move box invisible after tracking has ended
                        DrawingFrame.toggleTracking(drawingFrame, false);
                    });
                    allDrawingMoveBoxes = [];
                } else {
                    // make move box invisible after tracking has ended
                    DrawingFrame.toggleTracking(drawingFrame, false);
                    moveBox = null;
                }
            }

            // return the actual drawingTrackingHandler() method
            return function (event) {

                switch (event.type) {
                case 'tracking:start':
                    initializeTracking(event);
                    event.preventDefault(); // prevent native scrolling on touch devices
                    break;
                case 'tracking:move':
                    updateDrawingPosition(event);
                    event.preventDefault(); // prevent native scrolling on touch devices
                    break;
                case 'tracking:scroll':
                    self.scrollRelative(event.scrollX, event.scrollY);
                    updateDrawingPosition(event);
                    break;
                case 'tracking:end':
                    updateDrawingPosition(event);
                    finalizeDrawingTracking(event, true);
                    break;
                case 'tracking:cancel':
                    finalizeDrawingTracking(event, false);
                    break;
                }
            };
        }()); // end of drawingTrackingHandler() local scope

        // --------------------------------------------------------------------

        /**
         * Handles all tracking events in this grid pane.
         */
        function trackingStartHandler(event) {

            var // target node
                targetNode = $(event.target);

            // starts tracking with the specified type and event handler
            function startTracking(type, trackingHandler, options) {

                // store the tracking type
                trackingType = type;
                // store whether tracking in read-only mode is supported
                readOnlyTracking = Utils.getBooleanOption(options, 'readOnlyTracking', false);

                // handle all tracking events for this tracking cycle
                PaneUtils.processTrackingCycle(event, trackingHandler, function () {
                    lastTrackingType = trackingType;
                    trackingType = null;
                    view.grabFocus();
                });

                // Workaround for Firefox, it sends 'focusout' events after the old
                // selection DOM node that has been clicked is removed (the
                // 'focusout' event causes canceling the tracking sometimes).
                _.defer(function () { self.grabFocus(); });
            }

            // ignore inline pop-up menus, and other nodes that handle clicks internally
            if (targetNode.closest('.inline-popup,.skip-tracking').length > 0) {
                $.cancelTracking();
                return;
            }

            // select tracking mode according to event target node
            if (targetNode.closest('.drawing-layer ' + DrawingFrame.NODE_SELECTOR).length > 0) {
                // drawing layer: select, move, and resize drawing objects
                startTracking('drawing', drawingTrackingHandler);
            } else if (targetNode.closest('.highlight-layer [data-pos]').length > 0) {
                // highlight layer: move and resize highlighted cell ranges
                startTracking('highlight', highlightTrackingHandler);
            } else if (targetNode.closest('.selection-layer .autofill [data-pos]').length > 0) {
                // auto-fill: resize active selection range for auto-fill and auto-delete
                startTracking('autofill', autoFillTrackingHandler);
            } else if (targetNode.closest('.selection-layer .select [data-pos]').length > 0) {
                // change current selection range with resizer handles on touch devices
                startTracking('resizeselect', resizeSelectionTrackingHandler, { readOnlyTracking: true });
            } else if (event.trackingType === 'touch') {
                // touch devices: scroll, or set new selection
                startTracking('touchselect', touchSelectionTrackingHandler, { readOnlyTracking: true });
            } else if (lastTrackingType !== 'touchselect') {
                // set new selection, or extend current selection with mouse and modifier keys
                // (will be ignored on touch devices where a simple tap generates touch and mouse events)
                startTracking('mouseselect', mouseSelectionTrackingHandler, { readOnlyTracking: true });
            }
        }

        /**
         * Handles double clicks and starts the in-place edit mode.
         */
        function doubleClickHandler() {
            view.enterCellEditMode();
        }

        // initialization -----------------------------------------------------

        // initialize class members
        app.on('docs:init', function () {

            // the spreadsheet model and view instances
            model = app.getModel();
            view = app.getView();

            // leave in-place edit mode or cancel tracking, if document goes into read-only mode
            model.on('change:editmode', editModeHandler);

            // listen to view events
            view.on('change:activesheet', changeActiveSheetHandler);

            // tracking for cell selection with mouse/touch
            layerRootNode.on({
                'tracking:start': trackingStartHandler,
                dblclick: doubleClickHandler,
                'dragenter dragexit dragover dragleave': false
            });
        });

        // destroy all class members on destruction
        this.registerDestructor(function () {
            model = view = sheetModel = colCollection = rowCollection = drawingCollection = null;
        });

    } // class GridTrackingMixin

    // exports ================================================================

    return GridTrackingMixin;

});
